infl = read.csv("USInflation.csv",
                header = FALSE)$V1
y = infl
iT = length(infl)
x = c(1 : iT)

## 1 ############################################# 
## Polynomial regression

if (!require(splines)){
  install.packages("splines")
  library(splines)
}

## 1.1 Global polynomials

# the function bs() produces the basis function representation
# given in the slides (Set 5 Equation 1). The arguments are:
# degree (of the global or local polynomial) and df (degrees of freedom =
# degree + number of knots), or equivalently knots (number of knots)
polyfit2 = lm(y ~ bs(x = x, degree = 2)) 
summary(polyfit2)
polyfit3 = lm(y ~ bs(x = x, degree = 3)) 
summary(polyfit3)
polyfit4 = lm(y ~ bs(x = x, degree = 4)) 
summary(polyfit4)
## Without using any knots in the argument of the function
## bs(), the regression is global.

plot(x, y, type = "l", main = "US Inflation")
lines(x, polyfit2$fit, col = "red", lwd = 2)
lines(x, polyfit3$fit, col = "blue", lwd = 2)
lines(x, polyfit4$fit, col = "green", lwd = 2)
legend("topright",
       legend = c("Original series", "Degree = 2", "Degree = 3", "Degree = 4"),
       col = c("black", "red", "blue", "green"), lty = 1)

## The next plots show us that global polynomials could
## behave bad locally (especially at the extremes of the
## sample)
par(mfrow = c(2, 2))
plot(x, hatvalues(polyfit2), type = "p") # hatvalues() extracts the residuals
plot(x, hatvalues(polyfit3), type = "p")
plot(x, hatvalues(polyfit4), type = "p")
par(mfrow = c(1, 1))

## 1.2 Local polynomials (or splines)
## We need to decide the number of knots: Akaike information criterion (AIC)
choice_knots = data.frame(knots = c(1 : 50), AIC = rep(NA, 50))
for(i in 1:50){
  splinefit3 = lm(y ~ bs(x = x, degree = 3, df = 4 + i))
  choice_knots[i, 2] = extractAIC(splinefit3)[2]
} 

which.min(choice_knots$AIC)
# 48 df (44 knots) refers to having one knot every 
# 18 observations

splinefit3_aic = lm(y ~ bs(x, degree = 3, df = 4 + 44))
summary(splinefit3_aic)


# Despite the number of knots (higher degrees of freedom, higher flexibility) 
# it could still not well approximate the extremes of the interval
par(mfrow = c(2, 1))
plot(x, y, type = "line", main = "Cubic spline, Knots = 44")
lines(x, splinefit3_aic$fit, col = "red", lwd = 2)
plot(x = x, y = hatvalues(splinefit3_aic), type = "p")
par(mfrow = c(1, 1))

# We regularize the problem by imposing linearity outside the extreme knots.
# The function ns() generates the basis function for natural cubic splines.
# We can still decide the number of knots by AIC.
natural_spline_fit = lm(y ~ ns(x, df = 2 + 44))

par(mfrow = c(2, 2))
plot(x, y, type = "line", main = "Natural cubic spline, Knots = 44")
lines(x, natural_spline_fit$fit, col = "red", lwd = 2)
plot(x, y, type = "line", main = "Cubic spline, Knots = 44")
lines(x, splinefit3_aic$fit, col = "red", lwd = 2)
plot(x = x, y = hatvalues(natural_spline_fit), type = "p")
plot(x = x, y = hatvalues(splinefit3_aic), type = "p")
par(mfrow = c(1, 1))

## Fitting splines entails the coice of the number of knots
## and their position: difficult choices.
## Solution: SMOOTHING SPLINES. These only require lambda to be tuned. 
sspline_fit_cv = smooth.spline(x = x, y = y, cv = T)  

## In the next two lines we get the degrees of freedom
## of the smoothing spline (trace of hat matrix)
df_ss = sspline_fit_cv$df

par(mfrow = c(2, 1))
plot(hatvalues(sspline_fit_cv), main = "Smoothing spline, df = 141")
plot(x = x, y = hatvalues(natural_spline_fit), type = "p", main = "Natural cubic spline, df = 44")
par(mfrow = c(1, 1))

plot(x, y, type = "l", main = "US Inflation")
lines(x, polyfit3$fit, col = "yellow3", lwd = 2)
lines(x, splinefit3_aic$fit, col = "blue", lwd = 2)
lines(x, natural_spline_fit$fit, col = "green", lwd = 2)
lines(x, fitted(sspline_fit_cv), col = "red", lwd = 2.5)
legend("topright",
       legend = c("Original series", "Global cubic", "Cubic spline (local)", "Natural cubic spline", "Smoothing spline"),
       col = c("black", "yellow3", "blue", "green", "red"), lty = 1)

## 2. 

## Kernel regression (Nadaraya - Watson)

# Bandwidth selection by leave one out cross validation (LOOCV)
cv_ksmooth <- function(x, y, h) {
  N = length(y)
  cvscore = 0
  for (i in 1:N)
  {
    ymi = y[-i] 
    xmi = x[-i] 
    nw.fit = ksmooth(xmi, ymi, 'normal', bandwidth = h, x.points = x[i])
    yhat = nw.fit$y
    cvscore = cvscore + (y[i] - yhat)^2
  }
  return(cvscore)
}

h.values = seq(1, 100, 1) # bandwidth candidate values
m = length(h.values)
cv.scores = rep(NA, m)
for (r in 1:m) {
  cv.scores[r] = cv_ksmooth(x, y, h.values[r])
}
plot(h.values, cv.scores, ylab = "CV Error", xlab = "Bandwidth h", 
     main = "Bandwidth selection by CV")
h.cv = h.values[which.min(cv.scores)]

nw_fit = ksmooth(x = x, y = y, kernel = "normal", bandwidth = 2)

plot(x, y, main = "US Inflation and its trend", type = "l")
lines(nw_fit, col = "blue")
legend("topright", legend = "NW estimator",
       col = "blue", lty = 1)


# 3. Comparing the two models

# mse
mean((nw_fit$y - y)^2)
mean((sspline_fit_cv$y - y)^2)

